import sys
import csv
from collections import defaultdict
from vladiate.exceptions import ValidationException
from vladiate.validators import EmptyValidator
from vladiate import logs


class Vlad(object):

    def __init__(self, source, validators={}, default_validator=EmptyValidator):
        self.logger = logs.logger
        self.failures = defaultdict(lambda: defaultdict(list))
        self.missing_validators = None
        self.missing_fields = None
        self.source = source
        self.validators = validators or getattr(self, 'validators', {})

        self.validators.update({
            field: [default_validator()]
            for field, value in self.validators.iteritems() if not value
        })

    def _log_debug_failures(self):
        for field_name, field_failure in self.failures.iteritems():
            self.logger.debug("\nFailure on field: \"{}\":".format(field_name))
            for i, (row, errors) in enumerate(field_failure.iteritems()):
                self.logger.debug("  {}:{}".format(self.source, row))
                for error in errors:
                    self.logger.debug("    {}".format(error))

    def _log_validator_failures(self):
        for field_name, validators_list in self.validators.iteritems():
            for validator in validators_list:
                if validator.bad:
                    self.logger.error(
                        "  {} failed {} time(s) on field: '{}'".format(
                            validator.__class__.__name__, validator.fail_count,
                            field_name))
                    invalid = list(validator.bad)
                    shown = ["'{}'".format(field) for field in invalid[:99]]
                    hidden = ["'{}'".format(field) for field in invalid[99:]]
                    self.logger.error(
                        "    Invalid fields: [{}]".format(", ".join(shown)))
                    if hidden:
                        self.logger.error(
                            "    ({} more suppressed)".format(len(hidden)))

    def _log_missing_validators(self):
        self.logger.error("  Missing validators for:")
        self._log_missing(self.missing_validators)

    def _log_missing_fields(self):
        self.logger.error("  Missing expected fields:")
        self._log_missing(self.missing_fields)

    def _log_missing(self, missing_items):
        self.logger.error(
            "{}".format("\n".join(["    '{}': [],".format(
                field.encode('string-escape'))
                for field in sorted(missing_items)])))

    def validate(self):
        self.logger.info("\nValidating {}(source={})".format(
            self.__class__.__name__, self.source))
        reader = csv.DictReader(self.source.open())

        self.missing_validators = set(reader.fieldnames) - set(self.validators)
        if self.missing_validators:
            self.logger.info("\033[1;33m" + "Missing..." + "\033[0m")
            self._log_missing_validators()
            return False

        self.missing_fields = set(self.validators) - set(reader.fieldnames)
        if self.missing_fields:
            self.logger.info("\033[1;33m" + "Missing..." + "\033[0m")
            self._log_missing_fields()
            return False

        for line, row in enumerate(reader):
            for field_name, field in row.iteritems():
                for validator in self.validators[field_name]:
                    try:
                        validator.validate(field, row=row)
                    except ValidationException, e:
                        self.failures[field_name][line].append(e)
                        validator.fail_count += 1

        if self.failures:
            self.logger.info("\033[0;31m" + "Failed :(" + "\033[0m")
            self._log_debug_failures()
            self._log_validator_failures()
            return False
        else:
            self.logger.info("\033[0;32m" + "Passed! :)" + "\033[0m")
            return True



import logging

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("vlad_logger")
sh = logging.StreamHandler()
sh.setLevel(logging.INFO)
sh.setFormatter(logging.Formatter('%(message)s'))
logger.addHandler(sh)
logger.propagate = False



import io
import boto
from urlparse import urlparse
from StringIO import StringIO


class VladInput(object):
    ''' A generic input class '''

    def __init__(self):
        raise NotImplementedError

    def open(self):
        raise NotImplementedError

    def __repr__(self):
        raise NotImplementedError


class LocalFile(VladInput):
    ''' Read from a local file path '''

    def __init__(self, filename):
        self.filename = filename

    def open(self):
        return open(self.filename, 'r')

    def __repr__(self):
        return "{}('{}')".format(self.__class__.__name__, self.filename)


class S3File(VladInput):
    ''' Read from a file in S3 '''

    def __init__(self, path=None, bucket=None, key=None):
        if path and not any((bucket, key)):
            self.path = path
            parse_result = urlparse(path)
            self.bucket = parse_result.netloc
            self.key = parse_result.path
        elif all((bucket, key)):
            self.bucket = bucket
            self.key = key
            self.path = "s3://{}{}"
        else:
            raise ValueError(
                "Either 'path' argument or 'bucket' and 'key' argument must be set.")

    def open(self):
        s3 = boto.connect_s3()
        bucket = s3.get_bucket(self.bucket)
        key = bucket.new_key(self.key)
        contents = key.get_contents_as_string()
        ret = io.BytesIO(bytes(contents))
        return ret

    def __repr__(self):
        return "{}('{}')".format(self.__class__.__name__, self.path)


class String(VladInput):
    ''' Read a file from a string '''

    def __init__(self, string_input=None, string_io=None):
        self.string_io = string_io if string_io else StringIO(string_input)

    def open(self):
        return self.string_io

    def __repr__(self):
        return "{}('{}')".format(self.__class__.__name__, '...')



class ValidationException(Exception):
    ''' Thrown when validation fails '''
    pass

class BadValidatorException(Exception):
    ''' Thrown when a validator is bad '''
    def __init__(self, extra):
        message = 'Row does not contain the following unique_with fields: {}'
        super(BadValidatorException, self).__init__(message.format(extra))



import pytest

from ..validators import *
from ..exceptions import *


class FakeRow(object):
    ''' A row with side effects '''

    def __init__(self, fields):
        self.fields = fields

    def __getitem__(self, key):
        return self.fields[key].pop(0)

    def keys(self):
        return self.fields.keys()


@pytest.mark.parametrize('field', [('42'), ('42.0'), ('-42.0'), ])
def test_float_validator_works(field):
    FloatValidator().validate(field)


@pytest.mark.parametrize('field', [('foo'), (' '), ('7..0'), ('4,200'), ])
def test_float_validator_fails(field):
    with pytest.raises(ValidationException):
        FloatValidator().validate(field)


@pytest.mark.parametrize('field', [('42'), ])
def test_int_validator_works(field):
    IntValidator().validate(field)


@pytest.mark.parametrize('field', [
    ('foo'),
    (' '),
    ('7..0'),
    ('4,200'),
    ('42.0'),
    ('-42.0'),
])
def test_int_validator_fails(field):
    with pytest.raises(ValidationException):
        IntValidator().validate(field)


@pytest.mark.parametrize('field_set, field', [
    (['foo'], 'foo'),
    (['foo', 'bar'], 'foo'),
])
def test_set_validator_works(field_set, field):
    SetValidator(field_set).validate(field)


@pytest.mark.parametrize('field_set, field', [
    ([], 'bar'),
    (['foo'], 'bar'),
    (['foo', 'bar'], 'baz'),
])
def test_set_validator_fails(field_set, field):
    with pytest.raises(ValidationException):
        SetValidator(field_set).validate(field)


@pytest.mark.parametrize('fields, row, unique_with', [
    ([], {}, []),
    (['foo', 'bar'], {}, []),
    (['foo', 'foo'], FakeRow({'some_field': ['bar', 'baz']}), ['some_field']),
    (['foo', 'foo', 'foo'], FakeRow({
        'some_field': ['bar', 'baz', 'bar'],
        'some_other_field': ['baz', 'bar', 'bar']
    }), ['some_field', 'some_other_field']),
])
def test_unique_validator_works(fields, row, unique_with):
    v = UniqueValidator(unique_with=unique_with)
    for field in fields:
        v.validate(field, row)


@pytest.mark.parametrize('fields, row, unique_with, exception', [
    (['foo', 'bar', 'bar'], {}, [], ValidationException),
    (['foo', 'foo'], FakeRow({'some_field': ['bar', 'bar']}), ['some_field'],
        ValidationException),
    (['foo', 'foo'], FakeRow({'some_field': ['bar', 'bar']}), ['other_field'],
        BadValidatorException),
])
def test_unique_validator_fails(fields, row, unique_with, exception):
    with pytest.raises(exception):
        v = UniqueValidator(unique_with=unique_with)
        for field in fields:
            v.validate(field, row)


@pytest.mark.parametrize('pattern, field', [
    (r'foo.*', 'foo'),
    (r'foo.*', 'foobar'),
])
def test_regex_validator_works(pattern, field):
    RegexValidator(pattern).validate(field)


@pytest.mark.parametrize('pattern, field', [
    (r'foo.*', 'afoo'),
    (r'^$', 'foo'),
])
def test_regex_validator_fails(pattern, field):
    with pytest.raises(ValidationException):
        RegexValidator(pattern).validate(field)


def test_empty_validator_works():
    EmptyValidator().validate("")


def test_empty_validator_fails():
    with pytest.raises(ValidationException):
        EmptyValidator().validate("foo")


def test_ignore_validator():
    Ignore().validate("foo")



import pytest

from ..inputs import *
from ..validators import *
from ..vlad import *


@pytest.mark.parametrize('kwargs', [
    ({'path':'s3://some.bucket/some/s3/key.csv'}),
    ({'bucket':'some.bucket', 'key':'/some/s3/key.csv'}),
])
def test_s3_input_works(kwargs):
    S3File(**kwargs)


@pytest.mark.parametrize('kwargs', [
    ({}),
    ({'path':'s3://some.bucket/some/s3/key.csv', 'bucket':'some.bucket'}),
    ({'path':'s3://some.bucket/some/s3/key.csv', 'key':'/some/s3/key.csv'}),
    ({'bucket':'some.bucket'}),
    ({'key':'/some/s3/key.csv'}),
])
def test_s3_input_fails(kwargs):
    with pytest.raises(ValueError):
        S3File(**kwargs)


@pytest.mark.parametrize('kwargs', [
    ({'string_input':'ColA,ColB\n,'}),
    ({'string_io':StringIO('ColA,ColB\n,')}),
])
def test_string_input_works(kwargs):
    source = String(**kwargs)
    validators = {'ColA': [], 'ColB': []}
    assert Vlad(source=source, validators=validators).validate()



import pytest

from ..vlad import Vlad
from ..inputs import *
from ..validators import *


def test_initialize_vlad():
    source = LocalFile('vladiate/examples/vampires.csv')
    validators = {
        'Column A': [
            UniqueValidator()
        ],
        'Column B': [
            SetValidator(['Vampire', 'Not A Vampire'])
        ]
    }
    assert Vlad(source=source, validators=validators).validate()


def test_initialize_vlad_no_source():
    with pytest.raises(TypeError):
        Vlad().validate()


def test_unused_validator_fails_validation():
    source = LocalFile('vladiate/examples/vampires.csv')
    validators = {
        'Column A': [
            UniqueValidator()
        ],
        'Column B': [
            SetValidator(['Vampire', 'Not A Vampire'])
        ],
        'Column C': [
            FloatValidator()
        ]
    }

    assert not Vlad(source=source, validators=validators).validate()


def test_validators_in_class_variable_are_used():
    source = LocalFile('vladiate/examples/vampires.csv')
    class TestVlad(Vlad):
        validators = {
            'Column A': [
                UniqueValidator()
            ],
            'Column B': [
                SetValidator(['Vampire', 'Not A Vampire'])
            ]
        }

    assert TestVlad(source=source).validate()



# encoding: utf-8

import sys, os
from setuptools import setup, find_packages, Command
from setuptools.command.test import test as TestCommand

version = '0.0.8'


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ['-x', 'vladiate/test']

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


def readme():
    with open('README.rst') as f:
        return f.read()


setup(
    name='vladiate',
    version=version,
    description="Vladiate is a strict validation tool for CSV files",
    classifiers=[
        "Development Status :: 4 - Beta",
        'Environment :: Console',
        'Natural Language :: English',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.0',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        "Intended Audience :: Developers",
    ],
    keywords='validate CSV vampires',
    author='Dustin Ingram',
    author_email='github@dustingram.com',
    url='http://github.com/di/vladiate',
    license='MIT',
    long_description=readme(),
    packages=find_packages(exclude=['examples', 'tests']),
    include_package_data=True,
    zip_safe=False,
    install_requires=['boto'],
    tests_require=['pytest'],
    cmdclass={'test': PyTest},
    entry_points={
        'console_scripts': [
            'vladiate = vladiate.main:main',
        ]
    }, )



import vladiate
from vladiate import Vlad
from vladiate import logs

import os
import sys
import inspect
from optparse import OptionParser
from pkg_resources import get_distribution


def parse_options():
    """
    Handle command-line options with optparse.OptionParser.
    Return list of arguments, largely for use in `parse_arguments`.
    """

    # Initialize
    parser = OptionParser(
        usage="vladiate [options] [VladClass [VladClass2 ... ]]")

    parser.add_option(
        '-f', '--vladfile',
        dest='vladfile',
        default='vladfile',
        help=
        "Python module file to import, e.g. '../other.py'. Default: vladfile")

    # List vladiate commands found in loaded vladiate files/source files
    parser.add_option(
        '-l', '--list',
        action='store_true',
        dest='list_commands',
        default=False,
        help="Show list of possible vladiate classes and exit")

    # Version number (optparse gives you --version but we have to do it
    # ourselves to get -V too. sigh)
    parser.add_option(
        '-V', '--version',
        action='store_true',
        dest='show_version',
        default=False,
        help="show program's version number and exit")

    # Finalize
    # Return three-tuple of parser + the output from parse_args (opt obj, args)
    opts, args = parser.parse_args()
    return parser, opts, args


def is_vlad(tup):
    """
    Takes (name, object) tuple, returns True if it's a public Vlad subclass.
    """
    name, item = tup
    return bool(
        inspect.isclass(item) and issubclass(item, Vlad) and
        hasattr(item, "source") and getattr(item, "source") and
        hasattr(item, "validators") and not name.startswith('_'))


def find_vladfile(vladfile):
    """
    Attempt to locate a vladfile, either explicitly or by searching parent dirs.
    """
    # Obtain env value
    names = [vladfile]
    # Create .py version if necessary
    if not names[0].endswith('.py'):
        names += [names[0] + '.py']
    # Does the name contain path elements?
    if os.path.dirname(names[0]):
        # If so, expand home-directory markers and test for existence
        for name in names:
            expanded = os.path.expanduser(name)
            if os.path.exists(expanded):
                if name.endswith('.py') or _is_package(expanded):
                    return os.path.abspath(expanded)
    else:
        # Otherwise, start in cwd and work downwards towards filesystem root
        path = '.'
        # Stop before falling off root of filesystem (should be platform
        # agnostic)
        while os.path.split(os.path.abspath(path))[1]:
            for name in names:
                joined = os.path.join(path, name)
                if os.path.exists(joined):
                    if name.endswith('.py') or _is_package(joined):
                        return os.path.abspath(joined)
            path = os.path.join('..', path)
    # Implicit 'return None' if nothing was found


def load_vladfile(path):
    """
    Import given vladfile path and return (docstring, callables).
    Specifically, the vladfile's ``__doc__`` attribute (a string) and a
    dictionary of ``{'name': callable}`` containing all callables which pass
    the "is a vlad" test.
    """
    # Get directory and vladfile name
    directory, vladfile = os.path.split(path)
    # If the directory isn't in the PYTHONPATH, add it so our import will work
    added_to_path = False
    index = None
    if directory not in sys.path:
        sys.path.insert(0, directory)
        added_to_path = True
    # If the directory IS in the PYTHONPATH, move it to the front temporarily,
    # otherwise other vladfiles -- like vlads's own -- may scoop the intended
    # one.
    else:
        i = sys.path.index(directory)
        if i != 0:
            # Store index for later restoration
            index = i
            # Add to front, then remove from original position
            sys.path.insert(0, directory)
            del sys.path[i + 1]
    # Perform the import (trimming off the .py)
    imported = __import__(os.path.splitext(vladfile)[0])
    # Remove directory from path if we added it ourselves (just to be neat)
    if added_to_path:
        del sys.path[0]
    # Put back in original index if we moved it
    if index is not None:
        sys.path.insert(index + 1, directory)
        del sys.path[0]
    # Return our two-tuple
    vlads = dict(filter(is_vlad, vars(imported).items()))
    return imported.__doc__, vlads


def main():
    parser, options, arguments = parse_options()
    logger = logs.logger

    if options.show_version:
        print "Vladiate %s" % (get_distribution('vladiate').version, )
        sys.exit(0)

    vladfile = find_vladfile(options.vladfile)
    if not vladfile:
        logger.error(
            "Could not find any vladfile! Ensure file ends in '.py' and see --help for available options.")
        sys.exit(1)

    docstring, vlads = load_vladfile(vladfile)

    if options.list_commands:
        logger.info("Available vlads:")
        for name in vlads:
            logger.info("    " + name)
        sys.exit(0)

    if not vlads:
        logger.error("No vlad class found!")
        sys.exit(1)

    # make sure specified vlad exists
    if arguments:
        missing = set(arguments) - set(vlads.keys())
        if missing:
            logger.error("Unknown vlad(s): %s\n" % (", ".join(missing)))
            sys.exit(1)
        else:
            names = set(arguments) & set(vlads.keys())
            vlad_classes = [vlads[n] for n in names]
    else:
        vlad_classes = vlads.values()

    # validate all the vlads
    for vlad in vlad_classes:
        vlad(vlad.source, validators=vlad.validators).validate()


if __name__ == '__main__':
    main()
